@@ -85,6 +85,9 @@ struct _ScoutWindowPrivate
   gboolean search_in_calendars;
   gboolean search_case_sensitive;
   gchar* current_search_text;
+  GSList* search_exclude_list;
+  GSList* search_include_list;
+  GSList* search_list;
   GThread* search_thread;
   gboolean search_active;
   calendar_search_settings_t* calendar_search_settings;
@@ -603,31 +606,141 @@ _get_contact_avatar_icon(ScoutWindow* self, const gchar* abook_uid)
   return avatar_icon;
 }
 
+gboolean
+_find_in_text2(const gchar* current_search_text, const gchar* text, const gchar* src_text, gchar** prefix, gchar** suffix)
+{
+  gchar* pos = strstr(text, current_search_text);
+  int i = 0;
+  int max_length = 80;
+  if(pos!=NULL)
+  {
+    if((*prefix) == NULL && (*suffix) == NULL)
+    {
+      glong query_offset = g_utf8_pointer_to_offset(text, pos);
+      glong query_char_len = g_utf8_strlen(current_search_text, -1);
+      glong text_char_len = g_utf8_strlen(text, -1);
+      glong prefix_char_len = g_utf8_pointer_to_offset(text, pos);
+      glong prefix_size = (pos - text);
+      glong suffix_char_len = text_char_len - (query_char_len + query_offset);
+      glong suffix_size = strlen(text) - (strlen(current_search_text)+(pos - text));
+    
+      if(text_char_len <= max_length)
+      {
+	*prefix = g_strndup(src_text, pos - text);
+	*suffix = g_strndup(pos + strlen(current_search_text), suffix_size);
+      }
+      else
+      {
+	prefix_char_len = (max_length/2 - (query_char_len/2 + query_char_len%2));
+	suffix_char_len = (max_length/2 - query_char_len/2);
+	if((query_offset - prefix_char_len < 0))
+	{
+	  suffix_char_len += prefix_char_len - query_offset;
+	  prefix_char_len = query_offset;
+	}
+	else if((suffix_char_len + query_offset + query_char_len)>text_char_len)
+	{
+	  prefix_char_len += (suffix_char_len + query_offset + query_char_len) - text_char_len;
+	  suffix_char_len = text_char_len - (query_offset + query_char_len);
+	}
+	*prefix = g_strndup(g_utf8_offset_to_pointer(src_text, query_offset - prefix_char_len),
+			    pos - g_utf8_offset_to_pointer(text, query_offset - prefix_char_len));
+	*suffix = g_strndup(g_utf8_offset_to_pointer(src_text, query_offset + query_char_len),
+			    g_utf8_offset_to_pointer(text, 
+						     query_offset + query_char_len + suffix_char_len) - 
+			    g_utf8_offset_to_pointer(text,query_offset + query_char_len));
+      }
+    }
+    return TRUE;
+  }
+  return FALSE;
+}
 
 gboolean
-_find_in_text(ScoutWindow* self, const gchar* text0, gchar** prefix, gchar** suffix)
+_find_not_in_text(GSList* negative_list, const gchar* text)
+{
+  while(negative_list)
+  {
+    gchar* current_search_text = (gchar*)negative_list->data;
+    gchar* pos = strstr(text, current_search_text);
+    if(pos!=NULL)
+    {
+      return TRUE;
+    }
+    negative_list = negative_list->next;
+  }
+  return FALSE;
+}
+
+gboolean
+_find_from_search_list_in_text(GSList* search_list, 
+			       const gchar* text, 
+			       const gchar* src_text, 
+			       gchar** prefix, 
+			       gchar** suffix,
+			       gchar** match)
+{
+
+  while(search_list)
+  {
+    gchar* current_search_text = (gchar*)search_list->data;
+    if(_find_in_text2(current_search_text, text, src_text, prefix, suffix))
+    {
+      *match = g_strdup(current_search_text);
+      return TRUE;
+    }
+    search_list = search_list->next;
+  }
+  return FALSE;
+}
+
+gboolean
+_find_all_in_text(GSList* search_list, 
+		  const gchar* text, 
+		  const gchar* src_text, 
+		  gchar** prefix, 
+		  gchar** suffix,
+		  gchar** match)
+{
+  while(search_list)
+  {
+    gchar* current_search_text = (gchar*)search_list->data;
+    if(!_find_in_text2(current_search_text, text, src_text, prefix, suffix))
+    {
+      if(*match)
+      {
+	g_free(*match);
+	*match = NULL;
+      }
+      if(*prefix)
+      {
+	g_free(*prefix);
+	*prefix = NULL;
+      }
+      if(*suffix)
+      {
+	g_free(*suffix);
+	*suffix = NULL;
+      }
+      return FALSE;
+    }
+    else
+    {
+      if((*match) == NULL)
+	*match = g_strdup(current_search_text);
+    }
+    search_list = search_list->next;
+  }
+  return TRUE;
+}
+
+gboolean
+_find_in_text(ScoutWindow* self, const gchar* text0, gchar** prefix, gchar** suffix, gchar** match)
 {
   gchar* text = NULL;
   const gchar* faulty_char = NULL;
   if(text0 == NULL || !g_utf8_validate(text0, -1, &faulty_char))
   {
-    /*
-    g_warning("text invalid %s %c\n", text0, faulty_char);
-    gchar* new_try = g_strndup(text0, faulty_char - text0);
-    if(_find_in_text(self, new_try, prefix, suffix))
-    {
-      g_free(new_try);
-      return TRUE;
-    }
-    g_free(new_try);
-    new_try = g_strndup(faulty_char+1, strlen(faulty_char));
-    if(_find_in_text(self, new_try, prefix, suffix))
-    {
-      g_free(new_try);
-      return TRUE;
-    }
-    g_free(new_try);
-    */
     return FALSE;
   }
   if(!self->priv->search_case_sensitive)
@@ -638,50 +751,21 @@ _find_in_text(ScoutWindow* self, const gchar* text0, gchar** prefix, gchar** suf
   {
     text = g_strdup(text0);
   }
-  gchar* pos = strstr(text, self->priv->current_search_text);
-  int i = 0;
-  int max_length = 80;
-  if(pos!=NULL)
+  gboolean rc = FALSE;
+  if(g_slist_length(self->priv->search_exclude_list) == 0 ||
+     !_find_not_in_text(self->priv->search_exclude_list, text))
   {
-    glong query_offset = g_utf8_pointer_to_offset(text, pos);
-    glong query_char_len = g_utf8_strlen(self->priv->current_search_text, -1);
-    glong text_char_len = g_utf8_strlen(text, -1);
-    glong prefix_char_len = g_utf8_pointer_to_offset(text, pos);
-    glong prefix_size = (pos - text);
-    glong suffix_char_len = text_char_len - (query_char_len + query_offset);
-    glong suffix_size = strlen(text) - (strlen(self->priv->current_search_text)+(pos - text));
-    
-    if(text_char_len <= max_length)
-    {
-      *prefix = g_strndup(text0, pos - text);
-      *suffix = g_strndup(pos + strlen(self->priv->current_search_text), suffix_size);
-    }
-    else
-    {
-      prefix_char_len = (max_length/2 - (query_char_len/2 + query_char_len%2));
-      suffix_char_len = (max_length/2 - query_char_len/2);
-      if((query_offset - prefix_char_len < 0))
+    if(g_slist_length(self->priv->search_include_list) == 0 ||
+       _find_all_in_text(self->priv->search_include_list, text, text0, prefix, suffix, match))
+      if((*prefix) != NULL || 
+	 (g_slist_length(self->priv->search_list) != 0 &&
+	  _find_from_search_list_in_text(self->priv->search_list, text, text0, prefix, suffix, match)))
       {
-	suffix_char_len += prefix_char_len - query_offset;
-	prefix_char_len = query_offset;
+	rc = TRUE;
       }
-      else if((suffix_char_len + query_offset + query_char_len)>text_char_len)
-      {
-	prefix_char_len += (suffix_char_len + query_offset + query_char_len) - text_char_len;
-	suffix_char_len = text_char_len - (query_offset + query_char_len);
-      }
-      *prefix = g_strndup(g_utf8_offset_to_pointer(text0, query_offset - prefix_char_len),
-			  pos - g_utf8_offset_to_pointer(text, query_offset - prefix_char_len));
-      *suffix = g_strndup(g_utf8_offset_to_pointer(text0, query_offset + query_char_len),
-			  g_utf8_offset_to_pointer(text, 
-						   query_offset + query_char_len + suffix_char_len) - 
-			  g_utf8_offset_to_pointer(text,query_offset + query_char_len));
-    }
-    g_free(text);
-    return TRUE;
   }
   g_free(text);
-  return FALSE;
+  return rc;
 }
 
 struct _result_data_t
@@ -747,7 +831,6 @@ _search_in_events(ScoutWindow* self,
 		  GdkPixbuf* outgoing_icon,
 		  int search_type)
 {
-  
   RTComElIter* rtcom_iter = rtcom_el_get_events(self->priv->rt_el, query);
   gboolean it = (rtcom_iter != NULL) &&rtcom_el_iter_first(rtcom_iter);
   while(self->priv->search_active && it)
@@ -756,7 +839,8 @@ _search_in_events(ScoutWindow* self,
     rtcom_el_iter_get_values(rtcom_iter, "free-text", &free_text, NULL);
     gchar* prefix = NULL;
     gchar* suffix = NULL;
-    if(_find_in_text(self, free_text, &prefix, &suffix))
+    gchar* match = NULL;
+    if(_find_in_text(self, free_text, &prefix, &suffix, &match))
     {
       gchar* remote_uid = NULL;
       gchar* local_uid = NULL;
@@ -798,7 +882,7 @@ _search_in_events(ScoutWindow* self,
 				  label_text, 
 				  time_string,
 				  prefix,
-				  self->priv->current_search_text,
+				  match,
 				  suffix);
 	g_free(label_text);
 
@@ -833,6 +917,7 @@ _search_in_events(ScoutWindow* self,
     }
     g_free(prefix);
     g_free(suffix);
+    g_free(match);
     g_free(free_text);
     it = rtcom_el_iter_next(rtcom_iter);
   }
@@ -875,11 +960,11 @@ _find_and_insert_calendar_event(ScoutWindow* self, CComponent* component, int ca
 {
   gchar* prefix = NULL;
   gchar* suffix = NULL;
-
+  gchar* match = NULL;
   int event_type = component->getType();
-  if(_find_in_text(self, component->getSummary().c_str(), &prefix, &suffix) ||
+  if(_find_in_text(self, component->getSummary().c_str(), &prefix, &suffix, &match) ||
      (event_type != 3 && 
-      _find_in_text(self, component->getDescription().c_str(), &prefix, &suffix)))
+      _find_in_text(self, component->getDescription().c_str(), &prefix, &suffix, &match)))
   {
     gchar* title_text = NULL;
     const gchar* faulty_char = NULL;
@@ -898,7 +983,7 @@ _find_and_insert_calendar_event(ScoutWindow* self, CComponent* component, int ca
 			      title_text,
 			      time_string,
 			      prefix,
-			      self->priv->current_search_text,
+			      match,
 			      suffix);
     g_free(title_text);
     g_free(time_string);
@@ -929,6 +1014,7 @@ _find_and_insert_calendar_event(ScoutWindow* self, CComponent* component, int ca
   }
   g_free(prefix);
   g_free(suffix);
+  g_free(match);
 }
 
 static void
@@ -993,7 +1079,7 @@ search_in_calendar(ScoutWindow* self)
 						       limit,
 						       offset,
 						       error);
-      while(component_list.size()>0)
+      while(self->priv->search_active && component_list.size()>0)
       {
 	offset += component_list.size();
 	for(int i = 0;i<component_list.size();++i)
@@ -1019,21 +1105,21 @@ search_in_calendar(ScoutWindow* self)
 						       limit,
 						       offset,
 						       error);
-      while(component_list.size()>0)
+      while(self->priv->search_active && component_list.size()>0)
+      {
+	offset += component_list.size();
+	for(int i = 0;i<component_list.size();++i)
 	{
-	  offset += component_list.size();
-	  for(int i = 0;i<component_list.size();++i)
-	    {
-	      _find_and_insert_calendar_event(self, component_list[i], calendar_list[j]->getCalendarId());
-	      delete component_list[i];
-	    }
-	  component_list = calendar_list[j]->getComponents(2,
-							   start_time,
-							   end_time,
-							   limit,
-							   offset,
-							   error);
+	  _find_and_insert_calendar_event(self, component_list[i], calendar_list[j]->getCalendarId());
+	  delete component_list[i];
 	}
+	component_list = calendar_list[j]->getComponents(2,
+							 start_time,
+							 end_time,
+							 limit,
+							 offset,
+							 error);
+      }
     }
     if(self->priv->calendar_search_settings->search_birthdays)
     {
@@ -1044,7 +1130,7 @@ search_in_calendar(ScoutWindow* self)
 						       limit,
 						       offset,
 						       error);
-      while(component_list.size()>0)
+      while(self->priv->search_active && component_list.size()>0)
 	{
 	  offset += component_list.size();
 	  for(int i = 0;i<component_list.size();++i)
@@ -1086,41 +1172,43 @@ _search_and_insert_contact_attribute(ScoutWindow* self, OssoABookContact* contac
     const gchar* value = NULL;
     gchar* prefix = NULL;
     gchar* suffix = NULL;
+    gchar* match = NULL;
     while(values)
     {
       value = (const gchar*) values->data;
-    if(value && 
-       _find_in_text(self, value, &prefix, &suffix))
-    {
-      const gchar* abook_uid = osso_abook_contact_get_persistent_uid(contact);
-      GdkPixbuf* avatar_icon = NULL;
-      if(abook_uid)
+      if(value && 
+	 _find_in_text(self, value, &prefix, &suffix, &match))
       {
-	avatar_icon = _get_contact_avatar_icon(self, abook_uid);
+	const gchar* abook_uid = osso_abook_contact_get_persistent_uid(contact);
+	GdkPixbuf* avatar_icon = NULL;
+	if(abook_uid)
+	{
+	  avatar_icon = _get_contact_avatar_icon(self, abook_uid);
+	}
+	GdkPixbuf* attribute_icon = _get_contact_attribute_icon(self, e_contact_field);
+
+	gchar* text_name = 
+	  g_markup_printf_escaped(_get_contact_markup_template(), 
+				  osso_abook_contact_get_display_name(contact),
+				  prefix,
+				  match,
+				  suffix);
+
+	_add_result_entry(self,
+			  self->priv->search_category_contacts_icon,
+			  text_name,
+			  attribute_icon,
+			  avatar_icon,
+			  SEARCH_CATEGORY_CONTACT,
+			  osso_abook_contact_get_persistent_uid(contact),
+			  0);
+
+	g_free(prefix);
+	g_free(suffix);
+	g_free(match);
+	return true;
       }
-      GdkPixbuf* attribute_icon = _get_contact_attribute_icon(self, e_contact_field);
-
-      gchar* text_name = 
-	g_markup_printf_escaped(_get_contact_markup_template(), 
-				osso_abook_contact_get_display_name(contact),
-				prefix,
-				self->priv->current_search_text,
-				suffix);
-
-      _add_result_entry(self,
-			self->priv->search_category_contacts_icon,
-			text_name,
-			attribute_icon,
-			avatar_icon,
-			SEARCH_CATEGORY_CONTACT,
-			osso_abook_contact_get_persistent_uid(contact),
-			0);
-
-      g_free(prefix);
-      g_free(suffix);
-      return true;
-    }
-    values = values->next;
+      values = values->next;
     }
   }
   g_list_free(attributes);
@@ -1302,6 +1390,113 @@ _threaded_search_func(gpointer data)
   return NULL;
 }
 
+static void
+_clear_current_search_list(ScoutWindow* self)
+{
+  if(self->priv->search_exclude_list)
+  {
+    g_slist_free(self->priv->search_exclude_list);
+    self->priv->search_exclude_list = NULL;
+  }
+  if(self->priv->search_include_list)
+  {
+    g_slist_free(self->priv->search_include_list);
+    self->priv->search_include_list = NULL;
+  }
+  if(self->priv->search_list)
+  {
+    g_slist_free(self->priv->search_list);
+    self->priv->search_list = NULL;
+  }
+}
+
+static void
+_split_list(ScoutWindow* self)
+{
+  GRegex * regexp;
+  GMatchInfo* match_info;
+  GError* error = NULL;
+  regexp = g_regex_new("((\\+|-)?(([^\"\\s]+\\b)|\"([^\"]+)\"))", 
+		       G_REGEX_OPTIMIZE, G_REGEX_MATCH_NOTEMPTY, &error);
+  if(error)
+  {
+    g_warning("regexperror %s\n", error->message);
+  }
+  else
+  {
+    g_regex_match(regexp, self->priv->current_search_text, G_REGEX_MATCH_NOTEMPTY, &match_info);
+    while(g_match_info_matches(match_info))
+    {
+      int i;
+      gint match_count = g_match_info_get_match_count(match_info);
+      if(match_count>3)
+      {
+	gchar* ex_in = NULL;
+	gchar* text = NULL;
+	ex_in = g_match_info_fetch(match_info, 2);
+	if(match_count==6)
+	{
+	  text = g_match_info_fetch(match_info, 5);
+	}
+	else
+	{
+	  text = g_match_info_fetch(match_info, 4);
+	}
+	if(ex_in == NULL || strlen(ex_in) == 0)
+	{
+	  self->priv->search_list = g_slist_append(self->priv->search_list, g_strdup(text));
+	}
+	else
+	  if(ex_in[0] == '-')
+	    self->priv->search_exclude_list = g_slist_append(self->priv->search_exclude_list, g_strdup(text));
+	  else if(ex_in[0] == '+')
+	    self->priv->search_include_list = g_slist_append(self->priv->search_include_list, g_strdup(text));
+	g_free(ex_in);
+	g_free(text);
+      }
+      g_match_info_next (match_info, NULL);
+    }
+    g_match_info_free (match_info);
+    g_regex_unref (regexp);
+  }
+}
+
+static void
+_init_search_keyword_list(ScoutWindow* self)
+{
+  _clear_current_search_list(self);
+  _split_list(self);
+  /*
+  if(self->priv->current_search_text)
+  {
+    gchar** keyword_list = g_strsplit(self->priv->current_search_text,
+				      " ",
+				      0);
+    gchar** string_iter = keyword_list;
+    while(*string_iter)
+    {
+      if(*string_iter)
+      {
+	if((*string_iter)[0] == '+')
+	{
+	  self->priv->search_include_list = g_slist_append(self->priv->search_include_list, g_strdup(*string_iter + 1));
+	}
+	else if((*string_iter)[0] == '-')
+	{
+	  self->priv->search_exclude_list = g_slist_append(self->priv->search_exclude_list, g_strdup(*string_iter + 1));
+	}
+	else
+	{
+	  self->priv->search_list = g_slist_append(self->priv->search_list, g_strdup(*string_iter));
+	}
+      }
+      string_iter++;
+    }
+    g_strfreev(keyword_list);
+  }
+  */
+}
+
 static void
 _do_search(ScoutWindow* self)
 {
@@ -1319,6 +1514,7 @@ _do_search(ScoutWindow* self)
       else
       {
 	g_free(self->priv->current_search_text);
+	self->priv->current_search_text = NULL;
       }
     }
 
@@ -1326,6 +1522,8 @@ _do_search(ScoutWindow* self)
       self->priv->current_search_text = g_utf8_strdown(gtk_entry_get_text(entry), -1);
     else
       self->priv->current_search_text = g_strdup(gtk_entry_get_text(entry));
+
+    _init_search_keyword_list(self);
     gtk_list_store_clear(self->priv->result_list);
     hildon_button_set_text(HILDON_BUTTON(self->priv->search_button),
 			   dgettext("hildon-libs", "wdgt_bd_stop"),NULL);
@@ -1990,6 +2188,8 @@ scout_window_init(ScoutWindow* self)
   self->priv->search_thread = NULL;
   self->priv->search_active = FALSE;
   self->priv->current_search_text = NULL;
+  self->priv->search_include_list = NULL;
+  self->priv->search_exclude_list = NULL;
   self->priv->search_in_contacts = TRUE;
   self->priv->search_in_conversations = TRUE;
   self->priv->search_in_calendars = TRUE;
@@ -2023,6 +2223,8 @@ scout_window_init(ScoutWindow* self)
   gtk_box_pack_start(GTK_BOX(category_selection_box), conversation_search, FALSE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(category_selection_box), calendar_search, FALSE, FALSE, 0);
   self->priv->search_entry_field = hildon_entry_new(HILDON_SIZE_FINGER_HEIGHT);
+  hildon_gtk_entry_set_placeholder_text(GTK_ENTRY(self->priv->search_entry_field),
+					"word1 word2 \"a sentence\" +MUST_INCLUDE -MUST_NOT_INCLUDE");
   g_object_set(self->priv->search_entry_field,"hildon-input-mode",
                HILDON_GTK_INPUT_MODE_FULL, NULL);
   self->priv->result_list = gtk_list_store_new(7, 
